import json
import os
import pydoc
from abc import ABC, abstractmethod
from itertools import chain
from typing import Generic, TypeVar, Dict, Union

import numpy as np
import pandas as pd

from sampo.utilities.serializers import CUSTOM_FIELD_SERIALIZER, CUSTOM_FIELD_DESERIALIZER, CUSTOM_TYPE_SERIALIZER, \
    CUSTOM_TYPE_DESERIALIZER, default_ndarray_serializer, default_dataframe_serializer, default_ndarray_deserializer, \
    default_dataframe_deserializer, default_np_int_deserializer, default_np_int_serializer

# TODO: describe constants
T = TypeVar('T', str, dict, list, tuple, str, bool, None)
S = TypeVar('S', bound='Serializable')
SS = TypeVar('SS', bound='StrSerializable')
JS = TypeVar('JS', bound='JSONSerializable')
AJS = TypeVar('AJS', bound='AutoJSONSerializable')

TYPE_HINTS = '_serializable_type_hints'


# TODO: Implement PartialSerializable, which can't completely deserialize itself, and just returns rich info to parent

# TODO: describe the class (description, parameters)
class Serializable(ABC, Generic[T, S]):
    @property
    @abstractmethod
    def serializer_extension(self) -> str:
        return NotImplemented

    @abstractmethod
    def _serialize(self) -> T:
        """
        Converts all the meaningful information from this instance to a generic representation
        :return: A generic representation
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, representation: T) -> S:
        """
        Creates class instance from a representation
        :param representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    @abstractmethod
    def load(cls, folder_path: str, file_name: str) -> S:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        ...

    # TODO: describe the function (return type)
    def dump(self, folder_path: str, file_name: str):
        """
        Serializes object and saves it to file
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        """
        ...

    @classmethod
    def get_full_file_name(cls, folder_path: str, file_name: str):
        """
        Combines path to folder, file name and extension to get full file name
        :param folder_path: Path to folder
        :param file_name: File name without extension
        :return: Full file path, name and extension
        """
        return os.path.join(folder_path, f'{file_name}.{cls.serializer_extension}')


# TODO: describe the class (description, parameters)
class StrSerializable(Serializable[str, SS], ABC, Generic[SS]):
    serializer_extension: str = 'dat'

    @abstractmethod
    def _serialize(self) -> str:
        """
        Converts object to str representation
        :return: str representation of the object
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, str_representation: str) -> SS:
        """
        Creates class instance from a str representation
        :param str_representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    def load(cls, folder_path: str, file_name: str) -> SS:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        full_file_name = cls.get_full_file_name(folder_path, file_name)
        with open(full_file_name, 'r', encoding='utf-8') as read_file:
            str_representation = read_file.read()
        return cls._deserialize(str_representation)

    # TODO: describe the function (return type)
    def dump(self, folder_path: str, file_name: str) -> None:
        """
        Serializes object and saves it to file
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        """
        full_file_name = self.get_full_file_name(folder_path, file_name)
        serialized_str = self._serialize()
        with open(full_file_name, 'w', encoding='utf-8') as write_file:
            write_file.write(serialized_str)


# TODO: describe the class (description, parameters)
class JSONSerializable(Serializable[Dict[str,
                                         Union[
                                             dict,
                                             Union[list, tuple],
                                             str,
                                             Union[int, float],
                                             bool,
                                             None]],
                                    JS], ABC, Generic[JS]):
    serializer_extension: str = 'json'

    @abstractmethod
    def _serialize(self) -> T:
        """
        Converts all the meaningful information from this instance to a generic representation
        :return: A generic representation
        """
        ...

    @classmethod
    @abstractmethod
    def _deserialize(cls, representation: T) -> JS:
        """
        Creates class instance from a representation
        :param representation: Representation produced by _serialize method
        :return: New class instance
        """
        ...

    @classmethod
    def load(cls, folder_path: str, file_name: str) -> JS:
        """
        Factory method that produces a python object from the serialized version of it
        :param folder_path: Path to the folder, where the serialized file is saved
        :param file_name: File name without extension
        (the file extension should match with the one returned by serializer_extension method)
        :return: The constructed python object
        """
        full_file_name = cls.get_full_file_name(folder_path, file_name)
        with open(full_file_name, 'r', encoding='utf-8') as read_file:
            dict_representation = json.load(read_file)
        return cls._deserialize(dict_representation)

    # TODO: describe the function (return type)
    def dump(self, folder_path: str, file_name: str) -> None:
        """
        Serializes object and saves it to file
        :param folder_path: Path to the folder where the serialized file should be saved
        :param file_name: Name of the file without extension
        (the appended extension could be explored via serializer_extension method)
        """
        full_file_name = self.get_full_file_name(folder_path, file_name)
        serialized_dict = self._serialize()
        with open(full_file_name, 'w', encoding='utf-8') as write_file:
            json.dump(serialized_dict, write_file)


# TODO: Implement automotive Enum serialization/deserialization
# TODO: Implement auto serialization of typed collection
# TODO: describe the class (parameters)
class AutoJSONSerializable(JSONSerializable[AJS], ABC):
    """
    Parent class for serialization of classes, which can be automatically converted to dict with Serializable properties
    and custom (de-)serializers, marked with custom_serializer and custom_deserializer decorators.
    """
    serializer_extension: str = 'json'

    _default_serializers_deserializers = {
        'serializers':
            {
                str(np.ndarray): default_ndarray_serializer,
                str(pd.DataFrame): default_dataframe_serializer,
                str(np.int32): default_np_int_serializer
            },
        'deserializer':
            {
                str(np.ndarray): default_ndarray_deserializer,
                str(pd.DataFrame): default_dataframe_deserializer,
                str(np.int32): default_np_int_deserializer
            }
    }

    # TODO: describe the function (description, return type)
    @classmethod
    @property
    def _default_serializers(cls):
        return cls._default_serializers_deserializers['serializers']

    # TODO: describe the function (description, return type)
    @classmethod
    @property
    def _default_deserializers(cls):
        return cls._default_serializers_deserializers['deserializer']

    # TODO: describe the function (description, return type)
    @classmethod
    @property
    def ignored_fields(cls):
        return []

    def _serialize(self) -> Dict[str,
                                 Union[
                                     dict,
                                     Union[list, tuple],
                                     str,
                                     Union[int, float],
                                     bool,
                                     None]]:
        """
        Converts all the meaningful information from this instance to dict representation
        with values of eligible types only to be converted to JSON data structures:
            - list or tuple (to array)
            - str (to string)
            - int or float (to number)
            - True or False (to true or false)
            - None (to null)
            - dict with values of the listed types (to object)
        :return: dict representation of the object
        """
        simple_types = dict, list, tuple, str, int, float, bool, type(None)
        custom_field_serializers = dict(chain(*[[(field, attr) for field in getattr(attr, CUSTOM_FIELD_SERIALIZER)]
                                                for attr in [getattr(self, name) for name in dir(self)]
                                                if hasattr(attr, CUSTOM_FIELD_SERIALIZER)]))
        custom_type_serializers = dict(chain(*[[(__type, attr) for __type in getattr(attr, CUSTOM_TYPE_SERIALIZER)]
                                               for attr in [getattr(self, name) for name in dir(self)]
                                               if hasattr(attr, CUSTOM_TYPE_SERIALIZER)]))
        type_hints = {}

        # TODO: describe the function (description, parameters, return type)
        def serialize_field(name, value):
            if name in custom_field_serializers:
                return custom_field_serializers[name](value)

            t = type(value)
            str_t = str(t)
            name_t = t.__name__
            if t in custom_type_serializers:
                type_hints[name] = str_t
                return custom_type_serializers[t](value)
            if str_t in custom_type_serializers:
                type_hints[name] = str_t
                return custom_type_serializers[str_t](value)
            if name_t in custom_type_serializers:
                type_hints[name] = name_t
                return custom_type_serializers[name_t](value)
            if str_t in self._default_serializers:
                type_hints[name] = str_t
                return self._default_serializers[str_t](value)
            if issubclass(t, Serializable):
                type_hints[name] = pydoc.classname(t, '')
                return value._serialize()
            if t in simple_types:
                return value
            raise ValueError(f'Cannot serialize field {name}.'
                             f' Provide custom serializer for it (via decorator @custom_serializer)'
                             f' or make inherit {type(value)} from Serializable.')

        return dict({TYPE_HINTS: type_hints},
                    **{k: serialize_field(k, v) for k, v in self.__dict__.items() if k not in self.ignored_fields})

    @classmethod
    def _deserialize(cls, dict_representation: dict) -> AJS:
        """
        Creates class instance from a dict representation
        :param dict_representation: Representation produced by _serialize method
        :return: New class instance
        """
        custom_field_deserializers = dict(chain(*[[(field, attr) for field in getattr(attr, CUSTOM_FIELD_DESERIALIZER)]
                                                for attr in [getattr(cls, name) for name in dir(cls)]
                                                if hasattr(attr, CUSTOM_FIELD_DESERIALIZER)]))
        custom_type_deserializers = dict(chain(*[[(str(__type), attr) for __type in getattr(attr,
                                                                                          CUSTOM_TYPE_DESERIALIZER)]
                                               for attr in [getattr(cls, name) for name in dir(cls)]
                                               if hasattr(attr, CUSTOM_TYPE_DESERIALIZER)]))
        type_hints = {}
        if TYPE_HINTS in dict_representation:
            type_hints = dict_representation[TYPE_HINTS]
            del dict_representation[TYPE_HINTS]

        # TODO: describe the function (description, parameters, return type)
        def deserialize_field(name, value):
            if name in custom_field_deserializers:
                return custom_field_deserializers[name](value)
            if name in type_hints:
                __type = type_hints[name]
                if __type in custom_type_deserializers:
                    return custom_type_deserializers[__type](value)
                if __type in cls._default_deserializers:
                    return cls._default_deserializers[__type](value)
                c = pydoc.locate(type_hints[name])
                return c._deserialize(value)
            return value

        fields = dict({f: None for f in cls.ignored_fields},
                      **{k: deserialize_field(k, v)
                         for k, v in dict_representation.items()
                         if k not in cls.ignored_fields})

        self = cls.__new__(cls)
        try:
            self.__dict__ = fields
        except:
            for k, v in fields.items():
                object.__setattr__(self, k, v)
        return self



from ast import literal_eval
from datetime import datetime, timedelta
from collections import defaultdict
from enum import auto, Enum
from itertools import chain
from operator import itemgetter, attrgetter
from typing import Dict, List, Optional, Union, Tuple

import numpy as np
import plotly.graph_objects
from pandas import DataFrame, Timestamp
from matplotlib import pyplot as plt
from plotly import express as px

from sampo.schemas.schedule import ResourceSchedule, ScheduleWorkDict, Schedule
from sampo.utilities.visualization.base import VisualizationMode, visualize
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.visualization.work_graph import SIZE_LIMIT, DPI_LIMIT, DEFAULT_DPI

SPLITTER = '#'


class EmploymentFigType(Enum):
    Classic = auto()
    Grouped = auto()
    DateLabeled = auto()
    WorkLabeled = auto()


def resource_employment_fig(schedule: Union[DataFrame, Schedule],
                            fig_type: EmploymentFigType = EmploymentFigType.Classic,
                            vis_mode: VisualizationMode = VisualizationMode.ShowFig,
                            file_name: Optional[str] = None,
                            project_start: datetime = datetime(year=2020, month=1, day=1)) \
        -> Optional[Union[plt.Figure, plotly.graph_objects.Figure]]:

    graph_data = (convert_schedule_df(schedule, fig_type)
                  if isinstance(schedule, DataFrame)
                  else get_schedule_df(schedule.to_schedule_work_dict, fig_type, project_start)) \
        if fig_type in [EmploymentFigType.WorkLabeled, EmploymentFigType.DateLabeled] \
        else get_workers_intervals(schedule, fig_type is EmploymentFigType.Grouped)

    return create_employment_fig(graph_data, fig_type, vis_mode, file_name)


def create_employment_fig(resources: Union[DataFrame, ResourceSchedule],
                          fig_type: EmploymentFigType,
                          vis_mode: VisualizationMode,
                          file_name: Optional[str] = '') \
        -> Optional[Union[plt.Figure, plotly.graph_objects.Figure]]:
    """

    :param resources:
    :param fig_type:
    :param vis_mode:
    :param file_name:
    :return:
    """

    if fig_type in [EmploymentFigType.WorkLabeled, EmploymentFigType.DateLabeled]:
        assert isinstance(resources, DataFrame), \
            f'Wrong data format. Expected DataFrame, got {type(resources)} instead.'

        if fig_type == EmploymentFigType.WorkLabeled:
            fig = px.timeline(data_frame=resources,
                              title=f"{'Resource load by tasks - chart'}",
                              x_start='time_start', x_end='time_end',
                              y='resource',
                              color='count',
                              hover_data={'count': True, 'time_start': True, 'time_end': False, 'resource': True},
                              text='count',
                              color_continuous_scale=px.colors.sequential.Agsunset_r)
            fig.update_traces(textposition='inside')
            works = {}
            for _, row in resources.iterrows():
                if row['time_start'] not in works:
                    works[row['time_start']] = row['work']

            w_lst = sorted(works.items(), key=itemgetter(0))

            fig.update_xaxes(tickangle=75,
                             # category_orders={'time_start': [x[1] for x in w_lst]},
                             tickmode='array',
                             tickvals=[x[0] for x in w_lst],
                             ticktext=[x[1] for x in w_lst])
        else:
            fig = px.timeline(data_frame=resources,
                              x_start='time_start', x_end='time_end',
                              title=f"{'Resource load by days - chart'}",
                              y='resource',
                              color='count',
                              hover_data=['count'],
                              text='count',
                              color_continuous_scale=px.colors.sequential.Agsunset_r)
            fig.update_traces(textposition='inside')
        return visualize(fig, vis_mode, file_name)

    assert isinstance(resources, ResourceSchedule), \
        f'Wrong data format. Expected ResourceSchedule, got {type(resources)} instead.'

    # 1. tune plot parameters according to the data size
    proposed_height = max(10, int(len(resources) * 0.2))  # used, if size limit does not exceed
    size = (20, proposed_height if proposed_height <= SIZE_LIMIT else SIZE_LIMIT)

    increased_dpi = DEFAULT_DPI + abs((proposed_height - SIZE_LIMIT) / 200)  # used to increase readability in big figs
    dpi = DEFAULT_DPI if proposed_height <= SIZE_LIMIT else min((increased_dpi, DPI_LIMIT))

    # 2. plot the data
    fig, ax = plt.subplots(figsize=size, dpi=dpi)
    for name, worker_schedule in resources.items():
        for worker_time in worker_schedule:
            ax.barh(name, left=worker_time[0], width=worker_time[1] - worker_time[0])

    for i in range(len(resources) + 1):
        ax.axhline(i - 0.5, color='black')

    plt.tight_layout()
    return visualize(fig, vis_mode, file_name)


def convert_schedule_df(schedule: DataFrame, fig_type: EmploymentFigType) -> DataFrame:
    first_day = schedule['start'].min()
    first_day = datetime(year=first_day.year, month=first_day.month, day=first_day.day)

    if fig_type is EmploymentFigType.DateLabeled:
        total_days = (schedule['finish'].max() - first_day).days + 1
        resource_schedule: Dict[str, np.ndarray] = {}
        for _, item in schedule.iterrows():
            start: Timestamp = (item['start'] - first_day).days
            finish: Timestamp = (item['finish'] - first_day).days
            workers: Union[str, Dict[str, int]] = item['workers']
            resources: Dict[str, int] = literal_eval(workers) if isinstance(workers, str) else workers
            for name in resources:
                if name not in resource_schedule:
                    resource_schedule[name] = np.array([0] * total_days)
                resource_schedule[name][start:finish + 1] += resources[name]

        data = list(chain(*[[(timedelta(days=i), timedelta(days=i + 1), r, c) for i, c in enumerate(s) if c > 0]
                            for r, s in resource_schedule.items()]))
        result = DataFrame.from_records(data, columns=['time_start', 'time_end', 'resource', 'count'])
        result.loc[:, ['time_start', 'time_end']] += first_day
        return result

    if fig_type is EmploymentFigType.WorkLabeled:
        resource_schedule: Dict[str, List[Tuple[int, str, int]]] = {}
        schedule.index = list(range(schedule.shape[0]))

        i = 0
        for _, item in schedule.iterrows():
            workers: Union[str, Dict[str, int]] = item['workers']
            resources: Dict[str, int] = literal_eval(workers) if isinstance(workers, str) else workers
            if len(resources) == 0:
                continue
            w_name = item['task_name']
            for name in resources:
                if resources[name] > 0:
                    if name not in resource_schedule:
                        resource_schedule[name] = []
                    resource_schedule[name].append((i, w_name, resources[name]))
            i += 1

        data = list(chain(*[[(timedelta(days=i), timedelta(days=i + 1), w, r, c) for i, w, c in s]
                            for r, s in resource_schedule.items()]))
        result = DataFrame.from_records(data, columns=['time_start', 'time_end', 'work', 'resource', 'count'])
        result.loc[:, ['time_start', 'time_end']] += first_day
        return result


def get_schedule_df(schedule: ScheduleWorkDict, fig_type: EmploymentFigType, project_start: datetime) -> DataFrame:
    if fig_type == EmploymentFigType.DateLabeled:
        max_time = max(schedule.values(), key=attrgetter('finish_time')).finish_time + 1
        resource_schedule: Dict[str, List[int]] = {}
        for item in schedule.values():
            start: Time = item.start_time
            finish: Time = item.finish_time
            resources: List[Worker] = item.workers
            for worker in resources:
                if worker.name not in resource_schedule:
                    resource_schedule[worker.name] = np.array([0] * int(max_time))
                resource_schedule[worker.name][start:finish + 1] += worker.count

        data = list(chain(*[[(timedelta(days=i), timedelta(days=i + 1), r, c) for i, c in enumerate(s) if c > 0]
                            for r, s in resource_schedule.items()]))
        result = DataFrame.from_records(data, columns=['time_start', 'time_end', 'resource', 'count'])
        result.loc[:, ['time_start', 'time_end']] += project_start
        return result

    if fig_type == EmploymentFigType.WorkLabeled:
        resource_schedule: Dict[str, List[Tuple[int, str, int]]] = {}
        for i, (work, item) in enumerate(sorted(list(schedule.items()), key=lambda x: x[1].start_time)):
            resources: List[Worker] = item.workers
            w_name = item.name
            for worker in resources:
                if worker.count > 0:
                    if worker.name not in resource_schedule:
                        resource_schedule[worker.name] = []
                    resource_schedule[worker.name].append((i, w_name, worker.count))

        data = list(chain(*[[(timedelta(days=i), timedelta(days=i + 1), w, r, c) for i, w, c in s]
                            for r, s in resource_schedule.items()]))
        result = DataFrame.from_records(data, columns=['time_start', 'time_end', 'work', 'resource', 'count'])
        result.loc[:, ['time_start', 'time_end']] += project_start
        return result


def get_workers_intervals(schedule: Schedule, group_workers_by_specializations: bool = False) \
        -> ResourceSchedule:
    resources_time_intervals_dict: Dict[Time, Dict[str, int]] = defaultdict(lambda: defaultdict(int))
    for item in schedule.works:
        start: Time = item.start_time
        finish: Time = item.finish_time
        workers: List[Worker] = item.workers
        for worker in workers:
            resources_time_intervals_dict[start][worker.name] -= worker.count
            resources_time_intervals_dict[finish][worker.name] += worker.count

    min_resources: Dict[str, int] = defaultdict(int)
    for milestone, resources in resources_time_intervals_dict.items():
        resources_time_intervals_dict[milestone] = dict(resources)
        for name in resources:
            min_resources[name] = min(min_resources[name], resources[name])
    resources_time_intervals_dict = dict(resources_time_intervals_dict)

    resources_time_intervals: List[Time, Dict[str, int]] = list(resources_time_intervals_dict.items())
    resources_time_intervals = sorted(resources_time_intervals, key=itemgetter(0))
    workers_intervals = defaultdict(list)
    for index, (start, resources) in enumerate(resources_time_intervals[:-1]):
        finish = resources_time_intervals[index + 1]
        for name in resources:
            resources[name] -= min_resources[name]
            if group_workers_by_specializations:
                workers_intervals[name].append((start, finish))
            else:
                for w_index in range(resources[name]):
                    workers_intervals[f'{name}{SPLITTER}{w_index}'].append((start, finish))

    if group_workers_by_specializations:
        workers_intervals = dict(sorted(dict(workers_intervals).items(),
                                        key=lambda item: item[0]))
    else:
        workers_intervals = dict(sorted(dict(workers_intervals).items(),
                                        key=lambda item: (item[0].split(SPLITTER)[0], int(item[0].split(SPLITTER)[1]))))
    return workers_intervals



from typing import List, Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.utilities.base_opt import coordinate_descent


class CoordinateDescentResourceOptimizer(ResourceOptimizer):

    def __init__(self, one_dimension_optimizer: Callable[[int, int, Callable[[int], Time]], Time]):
        self.one_dimension_optimizer = one_dimension_optimizer

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: List[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[List[Worker]], Time]):

        def fitness(worker_count: np.ndarray):
            for worker_ind in range(len(worker_team)):
                worker_team[worker_ind].count = worker_count[worker_ind]
            return get_finish_time(worker_team)

        count_worker_team = coordinate_descent(down_border, up_border,
                                               self.one_dimension_optimizer,
                                               fitness,
                                               optimize_array)
        if optimize_array:
            for i in range(len(worker_team)):
                if optimize_array[i]:
                    worker_team[i].count = count_worker_team[i]
        else:
            for i in range(len(worker_team)):
                worker_team[i].count = count_worker_team[i]



from collections.abc import Iterable
from typing import List, Any, Union, Dict, TypeVar, Callable

T = TypeVar('T')


# Returns a generator which should flatten any heterogeneous iterable
def flatten(xs: Iterable[Union[List[T], T]]) -> Iterable[T]:
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


K = TypeVar('K')
V = TypeVar('V')


def build_index(items: Iterable[T], key_getter: Callable[[T], K], value_getter: Callable[[T], V] = lambda x: x) \
        -> Dict[K, V]:
    """
    :param items: an iterable to index
    :param key_getter: a function that should retrieve index key from item
    :param value_getter: a function that should retrieve index value from item
    :return: dictionary that represents built index given by `key_getter` function
    """
    return {key_getter(item): value_getter(item) for item in items}


def reverse_dictionary(dictionary: Dict[K, V]) -> Dict[V, K]:
    return {value: key for key, value in dictionary.items()}



from collections import defaultdict
from dataclasses import dataclass, field
from random import Random
from typing import List, Optional, Callable

from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.identifiable import Identifiable
from sampo.schemas.requirements import WorkerReq, EquipmentReq, MaterialReq, ConstructionObjectReq
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.resources import Worker
from sampo.utilities.serializers import custom_serializer


# TODO: describe the class (description, parameters)
@dataclass
class WorkUnit(AutoJSONSerializable['WorkUnit'], Identifiable):
    worker_reqs: List[WorkerReq] = field(default_factory=list)
    equipment_reqs: List[EquipmentReq] = field(default_factory=list)
    material_reqs: List[MaterialReq] = field(default_factory=list)
    object_reqs: List[ConstructionObjectReq] = field(default_factory=list)
    group: Optional[str] = "default"
    is_service_unit: Optional[bool] = False
    # TODO Remove optional
    volume: Optional[float] = 1
    volume_type: Optional[str] = "unit"

    # TODO: describe the function (description, parameters, return type)
    @custom_serializer('worker_reqs')
    def worker_reqs_serializer(self, value):
        return [wr._serialize() for wr in value]

    # TODO: describe the function (description, parameters, return type)
    @classmethod
    @custom_serializer('worker_reqs', deserializer=True)
    def worker_reqs_deserializer(cls, value):
        return [WorkerReq._deserialize(wr) for wr in value]

    # TODO: describe the function (description, parameters, return type)
    # TODO: move this logit to WorkTimeEstimator
    def estimate_static(self, worker_list: List[Worker], work_estimator: WorkTimeEstimator = None) -> Time:
        if work_estimator:
            # TODO What?? We are loosing much performance
            #  when processing worker names EACH estimation
            workers = {w.name.replace("_res_fact", ""): w.count for w in worker_list}
            work_time = work_estimator.estimate_time(self.name.split("_stage_")[0], self.volume, workers)
            if work_time > 0:
                return work_time

        return self._abstract_estimate(worker_list, get_static_by_worker)

    # TODO: describe the function (description, parameters, return type)
    def estimate_stochastic(self, worker_list: List[Worker], rand: Optional[Random] = None) -> Time:
        return self._abstract_estimate(worker_list, get_stochastic_by_worker, rand)

    # TODO: describe the function (description, parameters, return type)
    def _abstract_estimate(self, worker_list: List[Worker],
                           get_productivity: Callable[[Worker, Optional[Random]], float],
                           rand: Optional[Random] = None) -> Time:
        groups = defaultdict(list)
        for w in worker_list:
            groups[w.name].append(w)
        times = [Time(0)]  # if there are no requirements for the work, it is done instantly
        for req in self.worker_reqs:
            if req.min_count == 0:
                continue
            name = req.kind
            command = groups[name]
            worker_count = sum([worker.count for worker in command], 0)
            if worker_count < req.min_count:
                return Time.inf()
            productivity = sum([get_productivity(c, rand) for c in command], 0) / worker_count
            productivity *= communication_coefficient(worker_count, req.max_count)
            if productivity == 0:
                return Time.inf()
            times.append(Time(req.volume // productivity))
        return max(times)


# TODO: describe the function (description, parameters, return type)
def get_static_by_worker(w: Worker, _: Optional[Random] = None):
    return w.get_static_productivity()

# TODO: describe the function (description, parameters, return type)
def get_stochastic_by_worker(w: Worker, rand: Optional[Random] = None):
    return w.get_stochastic_productivity(rand)


# Function is chosen because it has a quadratic decrease in efficiency as the number of commands on the object
# increases, after the maximum number of commands begins to decrease in efficiency, and its growth rate depends on
# the maximum number of commands.
# sum(1 - ((x-1)^2 / max_groups^2), where x from 1 to groups_count
# TODO: describe the function (description, parameters, return type)
def communication_coefficient(groups_count: int, max_groups: int) -> float:
    n = groups_count
    m = max_groups
    return 1 / (6 * m ** 2) * (-2 * n ** 3 + 3 * n ** 2 + (6 * m ** 2 - 1) * n)
